require 'digest/sha1'

module UuAppObjectStore
  module Concurrency

    # Class encapsulating lock data
    class Lock
      attr_reader :hash, :lock_code, :key, :expiration, :locked_objects


      #
      # Creates new instance of Lock.
      # @param [String] lock_code Code of the lock
      # @param [String] duration Expiration time in seconds. Defaults 60.
      def initialize(lock_code, duration = 60)
        key = _generate_key
        @lock_code = lock_code
        @key= key
        @hash= Digest::SHA1.base64digest("#{lock_code}:#{@key}")
        @expiration= Time.now+duration
        @locked_objects = {}
        @valid = true
      end

      # Extends expiration by duration.
      # @param [String] duration Expiration time in seconds. Defaults 60.
      # @raise [UuAppObjectStore::Error::ExpiredLock] if the lock is invalid
      def extend_expiration(duration = 60)
        raise UuAppObjectStore::Error::ExpiredLock unless valid?
        @expiration += duration
        schema_data = {}
        daos = @locked_objects.values
        daos.each do |dao|
          dao.extend_lock(self)
        end
        nil
      end

      # Check if the lock is valid
      # @return [Boolean]
      def valid?
        @expiration > Time.now
      end

      # unlock all locked objects
      # @raise [UuAppObjectStore::Error::ExpiredLock] if the lock is invalid
      def unlock_all
        raise UuAppObjectStore::Error::ExpiredLock unless valid?
        schema_data = []
        @locked_objects.each do |_, dao|
          schema_data << dao
        end
        schema_data.each do |dao|
          dao.unlock(self)
        end
        @locked_objects = []
        nil
      end


      # @private
      def add_object(object, dao) #:nodoc:
        key = "#{dao.coll}:#{object[:awid]}:#{object[:_id]}"
        @locked_objects[key] = dao
      end

      # @private
      def remove_object(object, dao) # :nodoc:
        key = "#{dao.coll}:#{object[:awid]}:#{object[:_id]}"
        @locked_objects.delete(key)
      end

      private


      def _generate_key
        SecureRandom.hex(24)
      end

    end

  end
end
