require 'uu_app/persistence/mongo/client'
require 'uu_app/sub_app_store/persistence_dao'
require 'uu_app/persistence'
require 'lru_redux'
require 'uu_app/sub_app_store'


require_relative '../error/duplicate_key'
require_relative '../error/invalid_revision'
require_relative '../error/missing_revision'
require_relative '../error/noi_exceeded'
require_relative '../error/object_not_found'
require_relative '../error/objectstore_error'
require_relative '../error/unexpected_exception'
require_relative '../error/soi_exceeded'
require_relative '../error/expired_lock'
require_relative '../error/object_locked'

module UuAppObjectStore
  module MongoDB

    # DAO providing objectstore for uuObject (Hash structure with
    # additional system attributes).
    class UuObjectDao < UuApp::SubAppStore::PersistenceDao

      # Default page size to be used for paged queries.
      DEFAULT_PAGE_SIZE = 1000

      # Constants determining basic revision strategy for parallel access
      REVISION = 'REVISION'
      NONE = 'NONE'

      CACHE_SIZE = 100

      # @private
      # A step (in percent) after which the current number of instances is recalculated
      STEP = 10

      attr_reader :max_noi, :max_soi, :coll

      #
      # Creates new instance of DAO.
      # @param [String] conn_uri Connection URI to the storage.
      # @param [String] coll Collection to be used.
      # @param [Hash] opts Optional parameters (max_noi, max_soi)
      def initialize(conn_uri, coll, opts = {})
        @max_noi = opts.delete(:max_noi)
        @max_soi = opts.delete(:max_soi)
        @count_cache = LruRedux::ThreadSafeCache.new(CACHE_SIZE) if @max_noi
        @client = UuApp::Persistence::Mongo::Client.new(conn_uri, opts)
        @coll = coll.to_sym
        @client[@coll].indexes.create_one({"sys.hashCode" => 1}, {partial_filter_expression: {"sys.hashCode" => {:$exists => true}}})
      end

      # Store new uuObject.
      # @param [Hash] uu_object Instance of uuObject to be stored (can contain +:id+ or +_id+ attribute).
      # @raise [UuAppObjectStore::Error::DuplicateKey] if the unique index is violated.
      # @return [Hash] Stored uuObject.
      def insert_one(uu_object)
        uu_object = _process_hash(uu_object)
        _convert_id(uu_object)
        uu_object[:sys] = _create_stamps

        # check maximal allowed size of object
        _check_soi(uu_object)

        # inserting over limit is forbidden
        expected_change = 1
        _check_current_noi_before(expected_change, uu_object[:awid])

        begin
          status = @client[@coll].insert_one(uu_object)
        rescue Mongo::Error::OperationFailure => e
          if e.message =~ /E1100[01]/
            raise UuAppObjectStore::Error::DuplicateKey
          else
            raise UuAppObjectStore::Error::UnexpectedException
          end
        rescue Mongo::Error => e
          raise UuAppObjectStore::Error::UnexpectedException
        end
        if status.nil? || !status.successful? || status.inserted_id.nil?
          raise UuAppObjectStore::Error::UnexpectedException
        end

        # update current number of uu_objects
        _recalculate_current_noi_after(1, uu_object[:awid])

        # process result
        result = UuApp::Util::UniformHash.new
        result.merge!(uu_object)
        result.delete(:_id)
        result[:id] = status.inserted_id.to_s
        result
      end

      # Insert the provided documents into the collection. Custom ID has
      # to be part of uuObject under +:id+ or +:_id+ attribute (these
      # fields are not required).
      # @param [Array<Hash>] uu_object_list Array of uuObjects to be stored.
      # @param [Boolean] ordered A boolean specifying whether the mongod instance should perform an ordered or unordered insert. Defaults to true.
      # @return [UuApp::Persistence::BulkResult<Hash>] Bulk result set.
      def insert_many(uu_object_list, ordered = true)
        uu_object_list = _process_array(uu_object_list)
        stamps = _create_stamps
        uu_object_list.each do |uu_object|
          uu_object[:sys] = stamps
          _check_soi(uu_object)
        end
        awid = uu_object_list[0][:awid] #TODO raise some error? if nil? validate all awids?
        uu_object_list.each { |uu_object| _convert_id(uu_object) }
        expected_change = uu_object_list.count
        _check_current_noi_before(expected_change, awid)

        begin
          result = @client[@coll].insert_many(uu_object_list, ordered: ordered)
            #TODO is this correct error handling only unexpected exceptions?
        rescue Mongo::Error::BulkWriteError => e
          if e.result['n_inserted']==0
            inserted_ids = []
            uu_object_list = []
          else
            # update current number of uu_objects (some of the uu_objects from uu_object_list might be inserted)
            _recalculate_current_noi_after(e.result['n_inserted'], awid)
            # process error and return result accordingly
            if ordered
              inserted_ids = e.result['inserted_ids'][0..e.result['n_inserted']-1]
              uu_object_list = uu_object_list[0..e.result['n_inserted']-1]
              error = UuAppObjectStore::Error::UnexpectedException, e.result['writeErrors'][0]['errmsg']
            else
              # select inserted documents
              indexes = e.result['writeErrors'].map { |error| error['index'] }
              inserted_ids = e.result['inserted_ids']
              indexes.reverse_each { |index|
                inserted_ids.delete_at(index)
                uu_object_list.delete_at(index)
              }
              error = e.result['writeErrors'].map { |error|
                raised_error = error['errmsg'] =~ /E1100[01]/ ? UuAppObjectStore::Error::DuplicateKey.new : UuAppObjectStore::Error::UnexpectedException.new(error['errmsg'])
                {error: raised_error, index:error['index']} }
            end
          end

          return _create_bulk_result(uu_object_list, inserted_ids, error, false)
        rescue Mongo::Error => e
          raise UuAppObjectStore::Error::UnexpectedException
        end

        # update current number of uu_objects (all of them were inserted)

        # inserting over limit is forbidden
        _recalculate_current_noi_after(expected_change, awid)

        # process result and return
        _create_bulk_result(uu_object_list, result.inserted_ids, nil, true)
      end

      # Find and returns the first stored uuObjects matching given filter.
      # @param [Hash] filter Filter to be used.
      # @param [Hash] projection  The fields to include or exclude from each uuObject in the result set.
      # @return [Hash] Result set.
      def find_one(filter, projection = {})
        _convert_id(filter)
        res = _find_wrapper(filter, {projection: projection, limit: 1})
        res[0]
      end

      # Find set of stored uuObjects matching given filter.
      # @param [Hash] filter Filter to be used.
      # @param [Hash] page_info Pagination info.
      # @param [Hash] projection  The fields to include or exclude from each uuObject in the result set.
      # @param [Hash] sort Specifies the order in which the query returns matching documents.
      # @return [Hash] Result set.
      def find(filter, page_info = {}, sort = {}, projection = {})
        _convert_id(filter)
        _convert_id(sort)
        _convert_id(projection)

        page_info = {} if page_info.nil?
        page_index = page_info[:pageIndex] ? page_info[:pageIndex].to_i : 0
        page_size = page_info[:pageSize] ? page_info[:pageSize].to_i : self.class::DEFAULT_PAGE_SIZE
        total_size = nil
        result = _find_wrapper(filter, {projection: projection, sort: sort}) do |r|
          total_size = r.count
          r.skip(page_index * page_size).limit(page_size)
        end

        {
            itemList: result,
            pageInfo: {
                pageIndex: page_index,
                pageSize: page_size,
                total: total_size
            }
        }
      end

      # Find a single uu_object and updates it, returning the updated uu_object.
      # @param [Hash] filter Filter to be used.
      # @param [Hash] uu_object Attributes with updated values. Has to contain revision for REVISION based strategy.
      # @param [String] revision_strategy Revision strategy to be used. Defaults to 'REVISION'.
      # @param [UuAppObjectStore::Concurrency::Lock] lock Used lock during the update. Defaults to nil.
      # @raise [UuAppObjectStore::Error::ExpiredLock] if used lock expired.
      # @raise [UuAppObjectStore::Error::ObjectLocked] if object is locked by another lock.
      # @raise [UuAppObjectStore::Error::LockDoesNotExist] if object is not locked by given lock.
      # @return [Hash] Updated uu_object
      def find_one_and_update(filter, uu_object, revision_strategy = REVISION, lock = nil)

        # symbolize keys and clone inputs
        filter = _process_hash(filter)
        update = _process_hash(uu_object)

        # check revision strategy
        object_revision = nil
        if revision_strategy == REVISION
          raise UuAppObjectStore::Error::MissingRevision if update[:sys].nil? || update[:sys][:rev].nil?
          object_revision = update[:sys][:rev]
        end

        #TODO fix other modificators $set, $unset,
        # updating system information is not allowed
        update.delete_if { |k, _| key = k.to_s
        key == 'sys' || key.start_with?('sys.')
        }

        # update modification stamp
        update[:'sys.mts'] = Time.now

        #
        _convert_id(filter)
        _convert_id(update)


        # prepare update
        update = _prepare_update(update)

        update_filter = {"$and" => [filter]}
        update_filter["$and"] << {'sys.rev': object_revision} if revision_strategy == REVISION


        lock_part = if lock
                      {"sys.lockHash" => lock.hash, "sys.lockExpTs" => {"$gt" => Time.now}}
                    else
                      {"$or" => [{"sys.lockHash" => {"$exists" => false}}, {"sys.lockHash" => {"$exists" => true}, "sys.ts" => {"$lt" => Time.now}}]}
                    end
        update_filter["$and"] << lock_part
        # {$and:[{awid:"5855bd69c50ed33188afa0ef11111111", code:"101-1"}, {"sys.rev":0}, {$or: [{"sys.hash": {$exists:false}},{"sys.hash":"test"},{"sys.hash": {$exists:true}, "sys.ts": {$lt : 4}}]}]}
        begin
          result = @client[@coll].find_one_and_update(update_filter, update, :return_document => :after)
        rescue Mongo::Error::OperationFailure => e
          @client.log_error(e)
          if e.message =~ /E1100[01]/
            raise UuAppObjectStore::Error::DuplicateKey
          else
            raise UuAppObjectStore::Error::UnexpectedException
          end
        rescue Mongo::Error => e
          raise UuAppObjectStore::Error::UnexpectedException
        end
        unless result
          object = @client[@coll].find(filter).first
          raise UuAppObjectStore::Error::ObjectNotFound unless object
          _eval_exception(object, lock, revision_strategy, object_revision)
        end
        _convert_to_hash(result)
      end

      # Perform an aggregation on the collection entries.
      # @param [Array<Hash>] pipeline The aggregation pipeline.
      # @param [Hash] opts The aggregation options:
      # @option opts [Boolean] :allow_disk_use Set to true if disk usage is allowed during
      #   the aggregation.
      # @option opts [Fixnum] :batch_size The number of documents to return per batch.
      # @option opts [Fixnum] :max_time_ms The maximum amount of time in milliseconds to allow the
      #   aggregation to run.
      # @option opts [Boolean] :use_cursor Indicates whether the command will request that the server
      #   provide results using a cursor.
      # @option opts [Boolean] :bypass_document_validation Whether or
      #   not to skip document level validation.
      # @return [Array] The aggregation result.
      def aggregate(pipeline, opts = {})
        begin
          result = @client[@coll].aggregate(pipeline, opts)
          result.inject([]) do |r, o|
            r << _convert_to_hash(o)
            r
          end
        rescue Mongo::Error => e
          raise UuAppObjectStore::Error::UnexpectedException
        end
      end

      # Delete an uu_object.
      # @param [Hash] filter Filter to be used.
      # @param [UuAppObjectStore::Concurrency::Lock] lock Lock. Defaults to nil.
      # @raise [UuAppObjectStore::Error::ExpiredLock] if used lock expired.
      # @raise [UuAppObjectStore::Error::ObjectLocked] if object is locked by another lock.
      # @raise [UuAppObjectStore::Error::LockDoesNotExist] if object is not locked by given lock.
      # @return [void]
      def delete_one(filter, lock = nil)
        filter = _process_hash(filter)
        _convert_id(filter)

        delete_filter = {"$and" => [filter]}
        lock_part = if lock
                      {"sys.lockHash" => lock.hash, "sys.lockExpTs" => {"$gt" => Time.now}}
                    else
                      {"$or" => [{"sys.lockHash" => {"$exists" => false}}, {"sys.lockHash" => {"$exists" => true}, "sys.lockExpTs" => {"$lt" => Time.now}}]}
                    end
        delete_filter["$and"] << lock_part
        begin
          status = @client[@coll].delete_one(delete_filter)
        rescue Mongo::Error => e
          raise UuAppObjectStore::Error::UnexpectedException
        end
        if status.nil? || !status.successful?
          @client.log_error(StandardError.new('Mongo::Operation::Write::Delete::Result was either nil or not successful'))
          raise UuAppObjectStore::Error::UnexpectedException
        elsif status.n == 0
          list = @client[@coll].find(filter)
          object = list.first
          if object
            raise UuAppObjectStore::Error::ExpiredLock if object[:sys][:lockHash] == lock.hash
            if object[:sys][:lockHash]
              raise UuAppObjectStore::Error::ObjectLocked
            else
              raise UuAppObjectStore::Error::LockDoesNotExist
            end
          end
        else
          # number of instances changed
          _recalculate_current_noi_after(-1, filter[:awid])
        end

        nil
      end

      # Delete uu_objects according to provided filter.
      # @param [Hash] filter Filter to be used.
      # @return [void]
      def delete_many(filter)
        filter = _process_hash(filter)
        _convert_id(filter)

        begin
          status = @client[@coll].delete_many(filter)
        rescue Mongo::Error => e
          @client.log_error(e)
          raise UuAppObjectStore::Error::UnexpectedException
        end
        if status.nil? || !status.successful?
          @client.log_error(StandardError.new('Mongo::Operation::Write::Delete::Result was either nil or not successful'))
          raise UuAppObjectStore::Error::UnexpectedException
        end
        # number of instances changed
        _recalculate_current_noi_after(-status.n, filter[:awid])
        nil
      end

      # Get item list of distinct values for a specified field.
      # @param [String, Symbol] field_name Name of the field
      # @param [Hash] filter Filter to be used
      # @return [Hash]
      def distinct(field_name, filter = nil)
        begin
          result = @client[@coll].distinct(field_name, filter)
        rescue Mongo::Error => e
          @client.log_error(e)
          raise UuAppObjectStore::Error::UnexpectedException
        end
        {itemList: result}
      end

      # Create index on specified hash attribute (or multiple
      # attributes).
      # @param [Hash] keys A hash of field name/direction pairs
      # @param [Hash] opts Options for this index:
      # @option opts [String] :name Custom index name.
      # @option opts [Boolean] :unique (false) If true, this index will enforce
      #   a uniqueness constraint on that field.
      # @option opts [Fixnum] :expire_after If set, defines dumber of seconds to
      #   expire documents in the collection after.
      # @option opts [Boolean] :sparse Whether the index should be sparse or not.
      # @option opts [Boolean] :background (false) If true, the index will be built
      #   in the background (only available for server versions >= 1.3.2 )
      # @option opts [Boolean] :drop_dups (false) If creating a unique index on
      #   this collection, this option will keep the first document the database indexes
      #   and drop all subsequent documents with duplicate values on this field.
      # @return [void]
      def create_index(keys, opts = {})
        begin
          status = @client[@coll].indexes.create_one(keys, opts)
        rescue Mongo::Error => e
          @client.log_error(e)
          raise UuAppObjectStore::Error::UnexpectedException
        end
        if status.nil? || !status.successful?
          @client.log_error(StandardError.new('Mongo::Operation::Result  was either nil or not successful'))
          raise UuAppObjectStore::Error::UnexpectedException
        end
        nil
      end

      # Return list of existing indexes.
      # @return [Array] Index list.
      def get_indexes
        result = []
        begin
          @client[@coll].indexes.each do |index|
            result << _convert_to_hash(index)
          end
        rescue Mongo::Error => e
          @client.log_error(e)
          raise UuAppObjectStore::Error::UnexpectedException
        end
        {itemList: result}
      end

      # Drop an index by its name.
      # @param [String] name The name of the index.
      # @return [void]
      def drop_index(name)
        begin
          status = @client[@coll].indexes.drop_one(name)
        rescue Mongo::Error => e
          @client.log_error(e)
          if e.message =~ /index not found/
            return
          else
            raise UuAppObjectStore::Error::UnexpectedException
          end
        end
        if status.nil? || !status.successful?
          @client.log_error(StandardError.new('Mongo::Operation::Result  was either nil or not successful'))
          raise UuAppObjectStore::Error::UnexpectedException
        end
        nil
      end

      # Get a count of uuObjects in the collection.
      # @param [Hash] filter Filter to be used.
      # @return [Fixnum] Count of matching uuObjects
      def count(filter)
        begin
          count = @client[@coll].count(filter)
        rescue Mongo::Error => e
          @client.log_error(e)
          raise UuAppObjectStore::Error::UnexpectedException
        end
        count
      end


      # lock first object with lock
      # @param [Hash] filter Filter to be used.
      # @param [UuObjecStore::Concurrency::Lock] lock Lock to be used.
      # @raise [UuAppObjectStore::Error::ExpiredLock] if lock is expired
      # @raise [UuAppObjectStore::Error::ObjectNotFound] if object does not exist
      def lock_one(filter, lock)
        filter = _process_hash(filter)
        raise UuAppObjectStore::Error::ExpiredLock unless lock.valid?
        _convert_id(filter)
        uu_object = _find_wrapper(filter, {limit: 1}).first

        raise UuAppObjectStore::Error::ObjectNotFound unless uu_object

        #locking the object
        lock_filter = {"$and" => [filter]}
        lock_part = {"$or" => [{"sys.lockHash" => {"$exists" => false}}, {"sys.lockHash" => {"$exists" => true}, "sys.lockExpTs" => {"$lt" => Time.now}}]}
        lock_filter["$and"] << lock_part

        update = {'$set' => {'sys.lockHash' => lock.hash, 'sys.lockExpTs' => lock.expiration}}
        begin
          locked_object = @client[@coll].find_one_and_update(lock_filter, update, :return_document => :after)
        rescue Mongo::Error::OperationFailure => e
          @client.log_error(e)
          if e.message =~ /E1100[01]/
            raise UuAppObjectStore::Error::DuplicateKey
          else
            raise UuAppObjectStore::Error::UnexpectedException
          end
        rescue Mongo::Error => e
          raise UuAppObjectStore::Error::UnexpectedException
        end

        raise UuAppObjectStore::Error::ObjectLocked unless locked_object
        lock.add_object(locked_object, self)


        locked_object
      end

      # unlock first object with lock
      # @param [Hash] filter Filter to be used.
      # @param [UuObjecStore::Concurrency::Lock] lock Lock to be used.
      # @raise [UuAppObjectStore::Error::ExpiredLock] if used lock is expired
      def unlock_one(filter, lock)
        raise UuAppObjectStore::Error::ExpiredLock unless lock.valid?
        filter = _process_hash(filter)
        _convert_id(filter)
        filter.merge!({"sys.lockHash" => lock.hash, "sys.lockExpTs" => {"$gt" => Time.now}})

        uu_object = @client[@coll].find_one_and_update(filter, {'$unset' => {'sys.lockHash' => '', 'sys.lockExpTs' => ''}}, :return_document => :after)
        uu_object
        lock.remove_object(uu_object, self)
      end

      #@private
      def unlock(lock) # :nodoc:
        query = {'sys.lockHash' => lock.hash}
        @client[@coll].update_many(query, {'$unset' => {'sys.lockHash' => '', 'sys.lockExpTs' => ''}})
      end

      #@private
      def extend_lock(lock) # :nodoc:
        query = {'sys.lockHash' => lock.hash}
        @client[@coll].update_many(query, {'$set' => {'sys.lockExpTs' => lock.expiration}})
      end


      # Closes database connection.
      def close
        @client.close
      end

      #@private
      def clear_cache # :nodoc:
        @count_cache.clear if @count_cache
      end

      #@private
      def get_count_from_cache(awid) # :nodoc:
        @count_cache[awid] || 0 if @count_cache
      end

      private

      def _eval_exception(uu_object, lock, revision_strategy, object_revision)
        if revision_strategy == REVISION && uu_object[:sys][:rev] != object_revision
          raise UuAppObjectStore::Error::InvalidRevision
        elsif lock && uu_object[:sys][:lockHash] == lock.hash
          raise UuAppObjectStore::Error::ExpiredLock
        elsif (lock && uu_object[:sys][:lockHash] && uu_object[:sys][:lockHash] != lock.hash) || uu_object[:sys][:lockHash]
          raise UuAppObjectStore::Error::ObjectLocked
        elsif lock && uu_object[:sys][:lockHash].nil?
          raise UuAppObjectStore::Error::LockDoesNotExist
        else
          raise UuAppObjectStore::Error::ObjectStoreError
        end
      end

      # Check if size of uuObject to be stored isn't bigger than allowed.
      # @param [Hash] uu_object uuObject to be stored
      def _check_soi(uu_object)
        return unless @max_soi
        estimated_size = uu_object.to_bson.length
        # The _id (the key and its value) will be generated if it is not a part uu_object already.
        # It will add another 17 bytes to the size of the BSON.
        estimated_size += 17 unless uu_object.include?(:_id)
        raise UuAppObjectStore::Error::SoiExceeded.new("The size #{estimated_size} is bigger than maximal allowed byte size #{@max_soi}") if estimated_size > @max_soi
      end

      # Check if the total number of instances would be exceeded considering the change.
      # @param [Fixnum] change Expected change of the instance count
      # @param [String] awid Appworkspace id
      def _check_current_noi_before(change, awid)
        return unless @max_noi
        _load_item_count(awid) unless @count_cache.fetch(awid)
        raise UuAppObjectStore::Error::NoiExceeded.new 'Operation wasn\'t executed, it would cause exceeding of max. number of uu_objects in collection.' if change + (@count_cache[awid] || 0) > @max_noi
      end

      # Recalculate the number of instances in this collection. Fetches actual limit from database every 1/ 10 th of maximal
      # allowed count of uu_objects.
      # @param [Fixnum] change Expected change of the instance count
      def _recalculate_current_noi_after(change, awid)
        return unless @max_noi

        _load_item_count(awid) unless @count_cache.fetch(awid)

        actual_step = [(@max_noi - @count_cache[awid]) / STEP, 1].max
        before = @count_cache[awid]


        @count_cache[awid] += change

        steps_before = before / actual_step
        steps_after = @count_cache[awid] / actual_step

        if (steps_before < steps_after) || (@max_noi - @count_cache[awid] < 5) # recalculate
          @client.log_info("Fetching current number of instances from DB [#{before}, #{@count_cache[awid]}]",)
          _load_item_count(awid)
        end
      end

      def _load_item_count(awid)
        @count_cache[awid] = @client[@coll].count({awid: awid})
      end

      # Simple wrapper for find operation.
      # @param [Hash] filter Filter to be used.
      # @param [Hash] options Find options.
      # @return [Hash] Result set.
      # @private
      def _find_wrapper(filter, options)
        begin
          result = @client[@coll].find(filter, options)
          result = yield result if block_given?
          result.inject([]) do |r, o|
            r << _convert_to_hash(o)
            r
          end
        rescue Mongo::Error => e
          @client.log_error(e)
          raise UuAppObjectStore::Error::UnexpectedException
        end
      end

      # Create new set of uuObject system attributes.
      # @return [Hash] New uuObject system attributes.
      # @private
      def _create_stamps # :nodoc:
        stamps = Hash.new
        now = Time.now
        stamps[:cts] = now
        stamps[:mts] = now
        stamps[:rev] = 0
        stamps
      end

      # Convert +id+ or +_id+ key (if present) to BSON::ObjectId
      # if the conversion is possible.
      # @param [Hash] hash Hash to be processed.
      def _convert_id(hash)
        if hash.has_key?(:id) || hash.has_key?(:_id)
          id = nil
          id = hash.delete(:id) if hash.has_key?(:id)
          id = hash.delete(:_id) if hash.has_key?(:_id)
          if BSON::ObjectId.legal?(id)
            hash[:_id] = BSON::ObjectId.from_string(id)
          else
            hash[:_id] = id
          end
        end
      end

      # Create bulk result (merged with IDs)
      # @param [Array<Hash>] bulk
      # @param [Array<String>] inserted_ids
      # @param [Array<Hash>] errors
      # @param [Boolean] successful
      # @private
      def _create_bulk_result(bulk, inserted_ids, errors, successful)
        result = bulk.zip(inserted_ids).map do |res, id|
          ri = UuApp::Util::UniformHash.new
          ri.merge!(res)
          ri.delete(:_id)
          ri[:id] = id.to_s
          ri
        end
        result.inject(UuApp::Persistence::BulkResult.new(successful, errors)) do |r, o|
          r << _convert_to_hash(o)
          r
        end
      end

      # Convert given BSON object to Hash with symbolized names.
      # @param [BSON] bson BSON object.
      # @return [UuApp::Util::UniformHash] Hash with symbolized names.
      # @private
      def _convert_to_hash(bson) # :nodoc:
        return nil if bson.nil?
        bson.inject(UuApp::Util::UniformHash.new) do |r, (k, v)|
          if r.has_key?(k)
            e = UuAppObjectStore::Error::ObjectStoreException.new('Duplicate key in hash.')
            @client.log_error(e)
            raise e
          end
          if k =~ /^_id$/
            r[:id] = v.to_s
          elsif v.kind_of? Array
            r[k] = _process_array(v)
          elsif v.kind_of? Hash
            r[k] = _process_hash(v)
          else
            r[k] = v
          end
          r
        end
      end

      # Convert given Hash to Hash with symbolized names and all non-basic
      # values converted to String (using +to_s+ method).
      # @param [Hash] hash Hash to be processed.
      # @return [UuApp::Util::UniformHash] Hash with symbolized names.
      # @private
      def _process_hash(hash) # :nodoc:
        return nil if hash.nil?
        hash.inject(UuApp::Util::UniformHash.new) do |r, (k, v)|
          if r.has_key?(k)
            e = UuAppObjectStore::Error::ObjectStoreException.new('Duplicate key in hash.')
            @client.log_error(e)
            raise e
          end
          case v
            when nil, String, Numeric, TrueClass, FalseClass, Time, Date
              r[k] = v
            when Array
              r[k] = _process_array(v)
            when Hash
              r[k] = _process_hash(v)
            else
              r[k] = v.to_s
          end
          r
        end
      end

      # Iterate through array and converts all included
      # Hashes to Hashes with symbolized names.
      # @param [Array] array Array to be processed.
      # @return [Array] Array with symbolized Hashes.
      # @private
      def _process_array(array) # :nodoc:
        return nil if array.nil?
        array.inject([]) do |r, v|
          case v
            when nil, String, Numeric, TrueClass, FalseClass, Time, Date
              r << v
            when Array
              r << _process_array(v)
            when Hash
              r << _process_hash(v)
            else
              r << v.to_s
          end
          r
        end
      end

      def _prepare_update(update)

        prepared_update = Hash.new
        update.each do |k, v|
          if k.to_s.start_with?('$')
            prepared_update[k] = update.delete(k)
          end
        end
        # merge inc with increment of revision, if is contained in update query
        if prepared_update.keys.include?(:'$inc')
          prepared_update[:'$inc'].merge!('sys.rev': 1)
        else
          prepared_update[:'$inc'] = {'sys.rev': 1}
        end

        if prepared_update.keys.include?(:'$set')
          prepared_update[:'$set'].merge!(update)
        else
          prepared_update[:'$set'] = update
        end
        prepared_update[:'$set'] = update
        prepared_update
      end
    end # class
  end # module MongoDB
end # module UuAppObjectStore